# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports

# Third-party imports

# Local imports


class TestParentObject:

    def __str__(self):
        return 'This is ' + self.__class__.__name__


class TestChildObject(TestParentObject):
    pass


class TestOtherObject:

    def __str__(self):
        return 'This is ' + self.__class__.__name__



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from collections.abc import Hashable, Sequence, MutableSequence, Set, MutableSet

# Third-party imports
import pytest

# Local imports
from lookups import fixed
from .tools import TestParentObject, TestChildObject, TestOtherObject


obj = object()
obj2 = object()
parent = TestParentObject()
parent2 = TestParentObject()
child = TestChildObject()
child2 = TestChildObject()
other = TestOtherObject()

MEMBER_FIXTURES = [
    # 0 member
    ([], object, []),

    # 1 member
    # None
    pytest.param([None], object, [], marks=pytest.mark.xfail),
    # object
    ([obj], object, [obj]),
    ([obj], TestOtherObject, []),
    # TestParentObject
    ([parent], object, [parent]),
    ([parent], TestParentObject, [parent]),
    ([parent], TestChildObject, []),
    ([parent], TestOtherObject, []),
    # TestChildObject
    ([child], object, [child]),
    ([child], TestParentObject, [child]),
    ([child], TestChildObject, [child]),
    ([child], TestOtherObject, []),

    # 2 members
    # None
    pytest.param([None, None], object, [], marks=pytest.mark.xfail),
    pytest.param([obj, None], object, [obj], marks=pytest.mark.xfail),
    # object
    ([obj, parent, child, other], object, [obj, parent, child, other]),
    ([obj, obj, child, other], object, [obj, obj, child, other]),
    ([obj, obj2, child, other], object, [obj, obj2, child, other]),
    ([other, other, other, other], object, [other, other, other, other]),
    # TestParentObject
    ([obj, parent, child, other], TestParentObject, [parent, child]),
    ([parent, parent, child, other], TestParentObject, [parent, parent, child]),
    ([parent, parent2, child, other], TestParentObject, [parent, parent2, child]),
    ([other, other, other, other], TestParentObject, []),
    # TestChildObject
    ([obj, parent, child, other], TestChildObject, [child]),
    ([child, parent, child, other], TestChildObject, [child, child]),
    ([child, parent, child2, other], TestChildObject, [child, child2]),
    ([other, other, other, other], TestChildObject, []),
]


def check_all_instances(expected, all_instances):
    assert isinstance(all_instances, Sequence)
    assert not isinstance(all_instances, MutableSequence)
    assert len(all_instances) == len(expected)
    assert all_instances == expected


def check_item(expected, item):
    if expected is None:
        assert item is None
        return

    assert item is not None
    assert isinstance(item, Hashable)

    assert item.get_display_name()
    assert item.get_id()

    assert item.get_instance() is expected

    assert issubclass(item.get_type(), type(expected))


def test_instantiation():
    assert fixed()


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup(members, search, expected):
    expected_first = expected[0] if expected else None
    lookup = fixed(*members)

    assert lookup.lookup(search) == expected_first


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_item(members, search, expected):
    expected_first = expected[0] if expected else None
    lookup = fixed(*members)

    item = lookup.lookup_item(search)
    check_item(expected_first, item)
    assert item == lookup.lookup_item(search)


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_all(members, search, expected):
    expected = tuple(expected)
    lookup = fixed(*members)

    all_instances = lookup.lookup_all(search)
    check_all_instances(expected, all_instances)


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_result(members, search, expected):
    expected = tuple(expected)
    expected_classes = {type(instance) for instance in expected}
    lookup = fixed(*members)

    result = lookup.lookup_result(search)
    assert result

    all_classes = result.all_classes()
    assert isinstance(all_classes, Set)
    assert not isinstance(all_classes, MutableSet)
    assert len(all_classes) == len(expected_classes)
    assert all_classes == expected_classes

    all_instances = result.all_instances()
    check_all_instances(expected, all_instances)

    all_items = result.all_items()
    assert isinstance(all_items, Sequence)
    assert not isinstance(all_items, MutableSequence)
    assert len(all_items) == len(expected)
    for item, instance, again in zip(all_items, expected, result.all_items()):
        check_item(instance, item)
        assert item == again


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_listeners(members, search, expected):
    lookup = fixed(*members)

    result = lookup.lookup_result(search)

    def call_me_back(result):
        pass

    result.add_lookup_listener(call_me_back)
    result.remove_lookup_listener(call_me_back)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations  # noqa: F407

# System imports
import weakref
from abc import ABC, abstractmethod
from threading import RLock
from typing import Collection, Type, Optional

# Third-party imports

# Local imports
from .lookups import LookupItem
from .generic_lookup import Content, Pair


class InstanceContent(Content):
    '''API for an owner of a lookup to be able to add and remove object instances to it.
    add/set/remove methods also support an optional parameter `convertor` in order to only add
    lightweight objects to the content and return the real, possibly heavy objects, only when
    someone access their instance through a lookup.

    Typical usage:
    > ic = InstanceContent()
    > lookup = GenericLookup(ic)
    > ic.add(an_object)
    > ic.set(a_sequence_of_objects)
    '''

    def add(self, instance: object, convertor: Convertor = None) -> bool:
        '''
        Adds an instance to the lookup.

        If convertor is None:
            If instance already exists in the lookup (duplication is determined by object equality)
            then the new instance replaces the old one in the lookup but listener notifications are
            NOT delivered in such case.

        else:
            The instance argument is just a key, not the actual value to appear in the lookup.
            The value will be created on demand, later when it is really needed by calling convertor
            methods.

            This method is useful to delay creation of heavy weight objects. Instead just register
            lightweight key and a convertor.

            To remove registered object from lookup use remove() with the same arguments.

        :param instance: The instance.
        :param convertor: Convertor which postpone an instantiation. If None then the instance is
        registered directly.
        :return: True if the instance has been added for the first time or False if some other
        instance equal to this one already existed in the content
        '''
        if convertor is None:
            return self._add_pair(SimpleItem(instance))
        else:
            return self._add_pair(ConvertingItem(instance, convertor))

    def remove(self, instance: object, convertor: Convertor = None) -> None:
        '''
        Remove instance.

        :param instance: The instance.
        :param convertor: Convertor, for an instance added with one.
        '''
        if convertor is None:
            self._remove_pair(SimpleItem(instance))
        else:
            self._remove_pair(ConvertingItem(instance, convertor))

    def set(self, instances: Collection[object], convertor: Convertor = None) -> None:
        '''
        Changes all pairs in the lookup to new values.

        Converts collection of instances to collection of pairs.

        :param instances: The collection of (Item) objects.
        :param convertor: The convertor to use or None.
        '''
        pairs: Collection[Pair]

        if convertor is None:
            pairs = tuple(SimpleItem(instance) for instance in instances)
        else:
            pairs = tuple(ConvertingItem(instance, convertor) for instance in instances)

        self._set_pairs(pairs)


class Convertor(ABC):
    '''Convertor postpones an instantiation of an object.'''

    @abstractmethod
    def convert(self, obj: object) -> object:
        '''
        Convert obj to other object.

        There is no need to implement cache mechanism. It is provided by Item.get_instance() method
        itself. However the method can be called more than once because instance is held just by
        weak reference.

        :param obj: The registered object.
        :return: The object converted from this object.
        '''

    @abstractmethod
    def type(self, obj: object) -> Type[object]:
        '''
        Return type of converted object.

        Accessible via Item.get_type().

        :param obj: The registered object.
        :return: The class that will be produced from this object (class or superclass of
        convert(obj))
        '''

    @abstractmethod
    def id(self, obj: object) -> str:
        '''
        Computes the ID of the resulted object.

        Accessible via Item.getId().

        :param obj: The registered object.
        :return: The ID for the object.
        '''

    @abstractmethod
    def display_name(self, obj: object) -> str:
        '''
        The human presentable name for the object.

        Accessible via Item.getDisplayName().

        :param obj: The registered object.
        :return: The name representing the object for the user.
        '''


class SimpleItem(LookupItem, Pair):
    '''Instance of one item representing an object.'''


class ConvertingItem(Pair):
    '''Instance of one item convertible to an object.'''

    def __init__(self, key: object, convertor: Convertor) -> None:
        if key is None:
            raise ValueError('None cannot be a lookup member')

        self._key = key
        self._convertor = convertor
        self._ref: Optional[weakref.ReferenceType[object]] = None
        self._lock = RLock()

    def _get_converted(self) -> Optional[object]:
        if self._ref is None:
            return None
        else:
            return self._ref()

    def get_display_name(self) -> str:
        return self._convertor.display_name(self._key)

    def get_id(self) -> str:
        return self._convertor.id(self._key)

    def get_instance(self) -> Optional[object]:
        with self._lock:
            converted = self._get_converted()

            if converted is None:
                converted = self._convertor.convert(self._key)
                self._ref = weakref.ref(converted)

            return converted

    def get_type(self) -> Type[object]:
        converted = self._get_converted()

        if converted is None:
            return self._convertor.type(self._key)
        else:
            return type(converted)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self._key == other._key
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._key)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
'''
Static factory methods for creating common lookup implementations.
'''

# System imports
from typing import Sequence, AbstractSet, Type, Optional, Callable

# Third-party imports

# Local imports
from . import singleton as singleton_module, simple
from .lookup import Lookup, Item, Result


def singleton(member: object, id_: str = None) -> Lookup:
    return singleton_module.SingletonLookup(member, id_)


def fixed(*members: object) -> Lookup:
    if not members:
        return EmptyLookup()

    elif len(members) == 1:
        return singleton(members[0])

    else:
        return simple.SimpleLookup(*members)


class NoResult(Result):

    def add_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def remove_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def all_classes(self) -> AbstractSet[Type[object]]:
        return frozenset()

    def all_instances(self) -> Sequence[object]:
        return tuple()

    def all_items(self) -> Sequence[Item]:
        return tuple()


class EmptyLookup(Lookup):

    NO_RESULT = NoResult()

    def lookup(self, cls: Type[object]) -> Optional[object]:
        return None

    def lookup_result(self, cls: Type[object]) -> Result:
        return self.NO_RESULT


class LookupItem(Item):

    def __init__(self, instance: object, id_: str = None) -> None:
        if instance is None:
            raise ValueError('None cannot be a lookup member')

        self._instance = instance
        self._id = id_

    def get_display_name(self) -> str:
        return self.get_id()

    def get_id(self) -> str:
        if self._id is not None:
            return self._id
        else:
            return str(self._instance)

    def get_instance(self) -> Optional[object]:
        return self._instance

    def get_type(self) -> Type[object]:
        return type(self._instance)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self._instance == other.get_instance()
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._instance)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from typing import Iterable, Collection, Set, Type, Optional, Tuple
from weakref import WeakValueDictionary

# Third-party imports

# Local imports
from . import generic_lookup as GL


class SetStorage(GL.Storage):
    '''Storing Pairs in a set datastructure, for GenericLookup.'''

    def __init__(self) -> None:
        self._content: Collection[GL.Pair] = set()

        # Do not serialize
        self._results: WeakValueDictionary[Type[object], GL.GLResult] = WeakValueDictionary()

    def begin_transaction(self, ensure: int) -> GL.Transaction:
        return SetTransaction(ensure, self._content)

    def end_transaction(self, transaction: GL.Transaction) -> Iterable[GL.GLResult]:
        self._content, change_set = transaction.new_content(self._content)

        results_to_notify = set()
        for pair in change_set:
            cls = pair.get_type()
            for result_type, result in self._results.items():
                if issubclass(cls, result_type):
                    results_to_notify.add(result)
                    result.clear_cache()

        return results_to_notify

    def lookup(self, cls: Type[object]) -> Iterable[GL.Pair]:
        for pair in self._content:  # TODO: improve
            if issubclass(pair.get_type(), cls):
                yield pair

    def register_result(self, result: GL.GLResult) -> None:
        self._results[result._cls] = result

    def find_result(self, cls: Type[object]) -> Optional[GL.GLResult]:
        return self._results.get(cls, None)


class SetTransaction(GL.Transaction):

    def __init__(self, ensure: int, current_content: Collection[GL.Pair]) -> None:
        self._new_list = set(current_content)
        self._changed: Set[GL.Pair] = set()

    def new_content(self, prev: Collection[GL.Pair]) -> Tuple[Collection[GL.Pair], Set[GL.Pair]]:
        return self._new_list, self._changed

    def add(self, pair: GL.Pair) -> bool:
        not_present = pair not in self._new_list
        if not_present:
            self._changed.add(pair)
        self._new_list.add(pair)
        return not_present

    def remove(self, pair: GL.Pair) -> None:
        self._changed.add(pair)
        self._new_list.remove(pair)

    def set_all(self, pairs: Collection[GL.Pair]) -> None:
        pairs = set(pairs)
        self._changed.update(pairs.symmetric_difference(self._new_list))
        self._new_list = pairs



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports

# Third-party imports

# Local imports
from .lookup import (
    Lookup, Result, Item,
    LookupProvider,
)
from .lookups import singleton, fixed
from .generic_lookup import GenericLookup, AbstractLookup
from .instance_content import InstanceContent, Convertor
from .__about__ import (
    __title__, __summary__, __uri__, __version__,
    __author__, __email__, __license__, __copyright__,
)

__all__ = [
    'Lookup', 'Result', 'Item',
    'LookupProvider',
    'GenericLookup', 'AbstractLookup', 'InstanceContent', 'Convertor',
    'singleton', 'fixed',

    '__title__', '__summary__', '__uri__', '__version__',
    '__author__', '__email__', '__license__', '__copyright__',
]



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
import gc
import threading
from collections.abc import Hashable, Sequence, MutableSequence, Set, MutableSet
from concurrent.futures import ThreadPoolExecutor

# Third-party imports
import pytest

# Local imports
from lookups import GenericLookup, InstanceContent, Convertor
from .tools import TestParentObject, TestChildObject, TestOtherObject


THREAD_WAITING_TIME = 0.025


obj = object()
obj2 = object()
parent = TestParentObject()
parent2 = TestParentObject()
child = TestChildObject()
child2 = TestChildObject()
other = TestOtherObject()


MEMBER_FIXTURES = [
    # 0 member
    ([], object, None),

    # 1 member
    pytest.param([None], object, None, marks=pytest.mark.xfail),
    # object
    ([obj], object, [obj]),
    ([obj], TestOtherObject, None),
    # TestParentObject
    ([parent], object, [parent]),
    ([parent], TestParentObject, [parent]),
    ([parent], TestChildObject, None),
    ([parent], TestOtherObject, None),
    # TestChildObject
    ([child], object, [child]),
    ([child], TestParentObject, [child]),
    ([child], TestChildObject, [child]),
    ([child], TestOtherObject, None),

    # 2 members
    # None
    pytest.param([None, None], object, None, marks=pytest.mark.xfail),
    pytest.param([obj, None], object, [obj], marks=pytest.mark.xfail),
    # object
    ([obj, parent, child, other], object, [obj, parent, child, other]),
    ([obj, obj, child, other], object, [obj, child, other]),
    ([obj, obj2, child, other], object, [obj, obj2, child, other]),
    ([other, other, other, other], object, [other]),
    # TestParentObject
    ([obj, parent, child, other], TestParentObject, [parent, child]),
    ([parent, parent, child, other], TestParentObject, [parent, child]),
    ([parent, parent2, child, other], TestParentObject, [parent, parent2, child]),
    ([other, other, other, other], TestParentObject, None),
    # TestChildObject
    ([obj, parent, child, other], TestChildObject, [child]),
    ([child, parent, child, other], TestChildObject, [child]),
    ([child, parent, child2, other], TestChildObject, [child, child2]),
    ([other, other, other, other], TestChildObject, None),
]


# Helpers


def setup_lookup(members, convertor=None):
    content = InstanceContent()
    lookup = GenericLookup(content)
    content.set(members, convertor=convertor)

    return content, lookup


def check_all_instances(expected, all_instances):
    assert isinstance(all_instances, Sequence)
    assert not isinstance(all_instances, MutableSequence)
    assert len(all_instances) == len(expected)
    for instance in all_instances:
        assert instance in expected
        expected.remove(instance)


def check_item(expected, item):
    if expected is None:
        assert item is None
        return

    assert item is not None
    assert isinstance(item, Hashable)

    assert item.get_display_name()
    assert item.get_id()

    assert item.get_instance() in expected
    idx = expected.index(item.get_instance())

    assert issubclass(item.get_type(), type(expected[idx]))
    return idx


# Objects life-cycle


def test_instantiation():
    assert InstanceContent()
    assert GenericLookup(InstanceContent())


def test_cannot_reuse_content():
    content, lookup = setup_lookup([])

    with pytest.raises(RuntimeError):
        GenericLookup(content)


# InstanceContent.add()


def test_add_before_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()

    for member in members:
        content.add(member)

    lookup = GenericLookup(content)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


def test_add_after_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)

    for member in members:
        content.add(member)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


# InstanceContent.set()


def test_set_before_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()

    content.set(members)

    lookup = GenericLookup(content)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


def test_set_after_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)

    content.set(members)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


# InstanceContent.remove()


def test_remove_before_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    content.set(members)

    for member in members:
        content.remove(member)

    lookup = GenericLookup(content)

    assert not lookup._storage
    # Can't test much more if no storage here...


def test_remove_after_attach():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)
    content.set(members)

    for member in members:
        try:
            content.remove(member)
        except KeyError:
            print('Could not remove', member)

    assert lookup._storage
    assert lookup._storage._content == set()

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert not check_in_storage(member)


# Basic lookup methods


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup(members, search, expected):
    content, lookup = setup_lookup(members)

    if expected:
        assert lookup.lookup(search) in expected
    else:
        assert lookup.lookup(search) is expected


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_item(members, search, expected):
    content, lookup = setup_lookup(members)

    item = lookup.lookup_item(search)
    check_item(expected, item)
    assert item == lookup.lookup_item(search)


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_all(members, search, expected):
    expected = list(expected) if expected is not None else []
    content, lookup = setup_lookup(members)

    all_instances = lookup.lookup_all(search)
    check_all_instances(expected, all_instances)


# Lookup.lookup_result()


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_lookup_result(members, search, expected):
    print(members, search, expected)
    expected = list(expected) if expected is not None else []
    print(members, search, expected)
    expected_classes = {type(instance) for instance in expected}
    content, lookup = setup_lookup(members)

    result = lookup.lookup_result(search)
    assert result

    all_classes = result.all_classes()
    assert isinstance(all_classes, Set)
    assert not isinstance(all_classes, MutableSet)
    assert len(all_classes) == len(expected_classes)
    assert all_classes == expected_classes

    all_instances = result.all_instances()
    check_all_instances(list(expected), all_instances)

    all_items = result.all_items()
    assert isinstance(all_items, Sequence)
    assert not isinstance(all_items, MutableSequence)
    assert len(all_items) == len(expected)
    for item, again in zip(all_items, result.all_items()):
        idx = check_item(expected, item)
        expected.pop(idx)
        assert item == again


def test_lookup_result_already_exist():
    content, lookup = setup_lookup([])

    result1 = lookup.lookup_result(object)
    result2 = lookup.lookup_result(object)

    assert result1 is result2


# Result listeners


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_listener(members, search, expected):
    print(members, search, expected)
    expected = list(expected) if expected is not None else []
    content, lookup = setup_lookup([])

    result = lookup.lookup_result(search)
    assert not result.all_items()

    def call_me_back(result):
        nonlocal called_with
        called_with = result
        print('Got called', result)

    called_with = None
    result.add_lookup_listener(call_me_back)

    # Adding members

    for member in members:
        print('Adding', member)
        added = content.add(member)
        if member in expected:
            if added:
                assert called_with
                assert member in result.all_instances()
                assert member in called_with.all_instances()
                called_with = None
            else:
                assert called_with is None
                assert member in result.all_instances()
        else:
            assert called_with is None

    # Removing members

    for member in members:
        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            if member in expected:
                assert called_with
                assert member not in result.all_instances()
                assert member not in called_with.all_instances()
                called_with = None
            else:
                assert called_with is None

    # Removing listener and adding/removing members

    result.remove_lookup_listener(call_me_back)

    for member in members:
        print('Adding', member)
        content.add(member)
        assert called_with is None

        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            assert called_with is None


def test_multiple_listeners():
    content, lookup = setup_lookup([])

    result_object = lookup.lookup_result(object)
    result_parent = lookup.lookup_result(TestParentObject)
    result_child = lookup.lookup_result(TestChildObject)
    result_other = lookup.lookup_result(TestOtherObject)

    def call_me_back(result):
        called_with[result._cls] = result
        print('Got called', result)

    called_with = {}
    result_object.add_lookup_listener(call_me_back)
    result_parent.add_lookup_listener(call_me_back)
    result_child.add_lookup_listener(call_me_back)
    result_other.add_lookup_listener(call_me_back)

    members = [obj, obj, obj2, parent]

    def check_for_a_class(member, added, cls, result_cls):
        if isinstance(member, cls):
            if added:
                assert cls in called_with
                assert called_with[cls] == result_cls
                del called_with[cls]
            else:
                assert cls not in called_with
        else:
            assert cls not in called_with

    # Adding members

    for member in members:
        print('Adding', member)
        added = content.add(member)
        check_for_a_class(member, added, object, result_object)
        check_for_a_class(member, added, TestParentObject, result_parent)
        check_for_a_class(member, added, TestChildObject, result_child)
        check_for_a_class(member, added, TestOtherObject, result_other)
        assert not called_with

    # Removing members

    for member in members:
        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            check_for_a_class(member, True, object, result_object)
            check_for_a_class(member, True, TestParentObject, result_parent)
            check_for_a_class(member, True, TestChildObject, result_child)
            check_for_a_class(member, True, TestOtherObject, result_other)
            assert not called_with

    # Removing listener and adding/removing members

    result_object.remove_lookup_listener(call_me_back)
    result_parent.remove_lookup_listener(call_me_back)
    result_child.remove_lookup_listener(call_me_back)
    result_other.remove_lookup_listener(call_me_back)

    for member in members:
        print('Adding', member)
        content.add(member)
        assert not called_with

        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            assert not called_with


@pytest.mark.xfail
def test_modify_lookup_from_listener():
    content, lookup = setup_lookup([])

    result_object = lookup.lookup_result(object)

    def call_me_back(result):
        content.add(obj2)

    result_object.add_lookup_listener(call_me_back)

    content.add(obj)


def test_del_result_clear_listener():
    content, lookup = setup_lookup([])
    result_object = lookup.lookup_result(object)

    def call_me_back(result):
        nonlocal called_with
        called_with = result
        print('Got called', result)

    called_with = None
    result_object.add_lookup_listener(call_me_back)

    content.add(obj)
    assert called_with is result_object
    called_with = None

    del result_object
    gc.collect()

    content.add(obj2)
    assert not called_with


# Result listeners with Executor


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_listener_with_executor(members, search, expected, request):
    print(members, search, expected)
    expected = list(expected) if expected is not None else []
    executor = ThreadPoolExecutor()
    request.addfinalizer(executor.shutdown)
    content = InstanceContent(notify_in=executor)
    lookup = GenericLookup(content)
    test_thread = threading.current_thread()
    called_event = threading.Event()

    result = lookup.lookup_result(search)
    assert not result.all_items()

    def call_me_back(result):
        nonlocal called_with, called_in_thread
        called_with = result
        called_in_thread = threading.current_thread()
        print('Got called', result)
        called_event.set()

    called_with = None
    called_in_thread = None
    result.add_lookup_listener(call_me_back)

    # Adding members

    for member in members:
        print('Adding', member)
        added = content.add(member)
        if member in expected:
            if added:
                assert called_event.wait(
                    THREAD_WAITING_TIME), 'Timeout waiting for callback to be called'
                called_event.clear()
                assert called_with
                assert member in result.all_instances()
                assert member in called_with.all_instances()
                assert called_in_thread
                assert called_in_thread != test_thread
                called_with = None
                called_in_thread = None
            else:
                assert not called_event.wait(
                    THREAD_WAITING_TIME), 'Callback got called when it should not'
                assert called_with is None
                assert called_in_thread is None
                assert member in result.all_instances()
        else:
            assert not called_event.wait(
                THREAD_WAITING_TIME), 'Callback got called when it should not'
            assert called_with is None
            assert called_in_thread is None

    # Removing members

    for member in members:
        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            if member in expected:
                assert called_event.wait(
                    THREAD_WAITING_TIME), 'Timeout waiting for callback to be called'
                called_event.clear()
                assert called_with
                assert member not in result.all_instances()
                assert member not in called_with.all_instances()
                assert called_in_thread
                assert called_in_thread != test_thread
                called_with = None
                called_in_thread = None
            else:
                assert not called_event.wait(
                    THREAD_WAITING_TIME), 'Callback got called when it should not'
                assert called_with is None
                assert called_in_thread is None

    # Removing listener and adding/removing members

    result.remove_lookup_listener(call_me_back)

    for member in members:
        print('Adding', member)
        content.add(member)
        assert called_with is None

        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            assert not called_event.wait(
                THREAD_WAITING_TIME), 'Callback got called when it should not'
            assert called_with is None
            assert called_in_thread is None


def test_multiple_listeners_with_executor(request):
    executor = ThreadPoolExecutor()
    request.addfinalizer(executor.shutdown)
    content = InstanceContent(notify_in=executor)
    lookup = GenericLookup(content)
    test_thread = threading.current_thread()

    result_object = lookup.lookup_result(object)
    result_parent = lookup.lookup_result(TestParentObject)
    result_child = lookup.lookup_result(TestChildObject)
    result_other = lookup.lookup_result(TestOtherObject)
    called_events = {
        object: threading.Event(),
        TestParentObject: threading.Event(),
        TestChildObject: threading.Event(),
        TestOtherObject: threading.Event(),
    }

    def call_me_back(result):
        called_with[result._cls] = result
        called_in_thread[result._cls] = threading.current_thread()
        print('Got called', result)
        called_events[result._cls].set()

    called_with = {}
    called_in_thread = {}
    result_object.add_lookup_listener(call_me_back)
    result_parent.add_lookup_listener(call_me_back)
    result_child.add_lookup_listener(call_me_back)
    result_other.add_lookup_listener(call_me_back)

    members = [obj, obj, obj2, parent]

    def check_for_a_class(member, added, cls, result_cls):
        if isinstance(member, cls):
            if added:
                assert called_events[cls].wait(
                    THREAD_WAITING_TIME), 'Timeout waiting for callback to be called'
                called_events[cls].clear()
                assert cls in called_with
                assert called_with[cls] == result_cls
                assert called_in_thread[cls]
                assert called_in_thread[cls] != test_thread
                del called_with[cls]
                del called_in_thread[cls]
            else:
                assert not called_events[cls].wait(
                    THREAD_WAITING_TIME), 'Callback got called when it should not'
                assert cls not in called_with
                assert cls not in called_in_thread
        else:
            assert not called_events[cls].wait(
                THREAD_WAITING_TIME), 'Callback got called when it should not'
            assert cls not in called_with
            assert cls not in called_in_thread

    # Adding members

    for member in members:
        print('Adding', member)
        added = content.add(member)
        check_for_a_class(member, added, object, result_object)
        check_for_a_class(member, added, TestParentObject, result_parent)
        check_for_a_class(member, added, TestChildObject, result_child)
        check_for_a_class(member, added, TestOtherObject, result_other)
        assert not called_with
        assert not called_in_thread

    # Removing members

    for member in members:
        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            check_for_a_class(member, True, object, result_object)
            check_for_a_class(member, True, TestParentObject, result_parent)
            check_for_a_class(member, True, TestChildObject, result_child)
            check_for_a_class(member, True, TestOtherObject, result_other)
            assert not called_with
            assert not called_in_thread

    # Removing listener and adding/removing members

    result_object.remove_lookup_listener(call_me_back)
    result_parent.remove_lookup_listener(call_me_back)
    result_child.remove_lookup_listener(call_me_back)
    result_other.remove_lookup_listener(call_me_back)

    for member in members:
        print('Adding', member)
        content.add(member)
        assert not called_with

        print('Removing', member)
        try:
            content.remove(member)
        except KeyError:
            continue
        else:
            assert not called_events[type(member)].wait(
                THREAD_WAITING_TIME), 'Callback got called when it should not'
            assert not called_with
            assert not called_in_thread


# InstanceContent Convertor


class MyConvertor(Convertor):

    def __init__(self, map_):
        self.map = map_
        self.convert_called = None
        self.type_called = None

    def reset(self):
        print('reset')
        self.convert_called = None
        self.type_called = None

    def convert(self, obj):
        print('convert called', obj)
        self.convert_called = obj
        return self.map[obj]

    def type(self, obj):
        print('type called', obj)
        self.type_called = obj
        return type(self.map[obj])

    def id(self, obj):
        return str(id(obj))

    def display_name(self, obj):
        return str(obj)


class KeyObject:

    def __repr__(self):
        return f'Key({super().__repr__()}) of {self.member!r}'


class WrapObject:

    def __init__(self, o):
        self.o = o

    def __hash__(self):
        return hash(self.o)

    def __eq__(self, other):
        return self.o == other

    def __repr__(self):
        return f'Wrap({super().__repr__()}) for {self.o!r}'


def make_convertor_maps(members):
    keys_to_members = {}
    members_to_keys = {}
    added_members = set()
    for member in members:
        if type(member) is object:
            member = WrapObject(member)
        if member in added_members:
            continue

        key = KeyObject()
        key.member = member
        keys_to_members[key] = member
        if member not in members_to_keys:
            members_to_keys[member] = []
        members_to_keys[member].append(key)
        added_members.add(member)

    return keys_to_members, members_to_keys


def test_convertor_none_key():
    with pytest.raises(ValueError):
        content, lookup = setup_lookup([None], MyConvertor({}))


def test_convertor_add():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)

    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)

    for key in keys_to_members.keys():
        content.add(key, convertor=convertor)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


def test_convertor_set():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)

    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)

    content.set(keys_to_members.keys(), convertor=convertor)

    assert lookup._storage
    assert lookup._storage._content

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert check_in_storage(member)


def test_convertor_remove():
    members = [object, parent, child, other, parent]
    content = InstanceContent()
    lookup = GenericLookup(content)

    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)
    content.set(keys_to_members.keys(), convertor=convertor)

    for key in keys_to_members.keys():
        try:
            content.remove(key, convertor=convertor)
        except KeyError:
            print('Could not remove', key)

    assert lookup._storage
    assert lookup._storage._content == set()

    def check_in_storage(obj):
        for pair in lookup._storage._content:
            if pair.get_instance() == obj:
                return True
        else:
            return False

    for member in members:
        assert not check_in_storage(member)


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_instance_convertor_lookup(members, search, expected):
    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)
    content, lookup = setup_lookup(keys_to_members.keys(), convertor)

    got = lookup.lookup(search)
    if expected:
        assert got in expected
        assert convertor.convert_called in members_to_keys[got]
        assert convertor.type_called in members_to_keys[got]
    else:
        assert got is None
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_instance_convertor_lookup_item(members, search, expected):
    def clear_cache():
        for item in lookup._storage._content:
            item._ref = None
        for result in lookup._storage._results.values():
            result.clear_cache()

    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)
    content, lookup = setup_lookup(keys_to_members.keys(), convertor)

    item = lookup.lookup_item(search)
    check_item(expected, item)
    # Clear, as we want to check just what lookup_item() does
    convertor.reset()
    clear_cache()
    assert item == lookup.lookup_item(search)
    assert convertor.convert_called is None, 'lookup_item() should not convert the instance'
    if expected:
        assert convertor.type_called in members_to_keys[item.get_instance()]
    else:
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_instance_convertor_lookup_all(members, search, expected):
    def clear_cache():
        for item in lookup._storage._content:
            item._ref = None
        for result in lookup._storage._results.values():
            result.clear_cache()

    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)
    content, lookup = setup_lookup(keys_to_members.keys(), convertor)

    expected = list(expected) if expected is not None else []
    all_instances = lookup.lookup_all(search)
    check_all_instances(list(expected), all_instances)
    # Clear, as we want to check just what lookup_all() does
    convertor.reset()
    clear_cache()
    assert all_instances == lookup.lookup_all(search)
    if expected:
        # Can't test much here
        assert isinstance(convertor.convert_called, KeyObject)
        assert isinstance(convertor.type_called, KeyObject)
    else:
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None


@pytest.mark.parametrize('members, search, expected', MEMBER_FIXTURES)
def test_instance_convertor_lookup_result(members, search, expected):
    def clear_cache():
        for item in lookup._storage._content:
            item._ref = None
        for result in lookup._storage._results.values():
            result.clear_cache()

    def swap_object_cls(cls):
        if cls is object:
            return WrapObject
        else:
            return cls

    expected = list(expected) if expected is not None else []
    expected_classes = {swap_object_cls(type(instance)) for instance in expected}
    keys_to_members, members_to_keys = make_convertor_maps(members)
    convertor = MyConvertor(keys_to_members)
    content, lookup = setup_lookup(keys_to_members.keys(), convertor)

    # lookup_result

    result = lookup.lookup_result(search)
    assert result
    assert convertor.convert_called is None, 'lookup_result() should not convert the instances'
    if expected:
        assert isinstance(convertor.type_called, KeyObject)
    else:
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None
    convertor.reset()
    clear_cache()

    # all_classes

    all_classes = result.all_classes()
    assert isinstance(all_classes, Set)
    assert not isinstance(all_classes, MutableSet)
    assert len(all_classes) == len(expected_classes)
    assert all_classes == expected_classes
    assert convertor.convert_called is None, 'all_classes() should not convert the instances'
    if expected:
        assert isinstance(convertor.type_called, KeyObject)
    else:
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None
    convertor.reset()
    clear_cache()

    # all_instances

    all_instances = result.all_instances()
    check_all_instances(list(expected), all_instances)
    if expected:
        # Can't test much here
        assert isinstance(convertor.convert_called, KeyObject)
        assert isinstance(convertor.type_called, KeyObject)
    else:
        assert convertor.convert_called is None
        if members:
            assert isinstance(convertor.type_called, KeyObject)
        else:
            assert convertor.type_called is None
    convertor.reset()
    clear_cache()

    # all_items

    all_items = result.all_items()
    assert isinstance(all_items, Sequence)
    assert not isinstance(all_items, MutableSequence)
    assert len(all_items) == len(expected)
    for item, again in zip(all_items, result.all_items()):
        idx = check_item(expected, item)
        expected.pop(idx)
        assert item == again
    # Clear, as we want to check just what lookup_all() does
    convertor.reset()
    clear_cache()
    assert all_items == result.all_items()
    assert convertor.convert_called is None, 'all_items() should not convert the instances'
    if members:
        assert isinstance(convertor.type_called, KeyObject)
    else:
        assert convertor.type_called is None


# GenericLookup subclass API


class MyLookup(GenericLookup):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initialise_called = None
        self._before_lookup_called = None

    def _initialise(self):
        self._initialise_called = True

    def _before_lookup(self, cls):
        self._before_lookup_called = cls


def test_subclass_api():
    content = InstanceContent()
    lookup = MyLookup(content)

    assert not lookup._initialise_called
    assert not lookup._before_lookup_called

    content.add(object())

    assert lookup._initialise_called
    assert not lookup._before_lookup_called

    lookup.lookup(object)

    assert lookup._before_lookup_called == object


# Miscelleanous features


def test_str():
    '''Checks that __str__ don\'t raise exceptions'''
    content, lookup = setup_lookup([])
    assert str(content)
    assert str(lookup)
    assert lookup._storage
    assert str(lookup._storage)
    assert str(lookup.lookup_result(object))

# TODO:
# Test GLResult caches



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
import importlib
import os

# Third-party imports
import pytest

# Local imports


def list_packages_in_folder(root_package_folder, ignores=None):
    if ignores is None:
        ignores = []

    basefolder = os.path.dirname(root_package_folder) + '/'

    def convert_folder_to_package(foldername):
        return foldername.replace(basefolder, '').replace('/', '.')

    for dirpath, dirnames, filenames in os.walk(root_package_folder):
        if '__init__.py' not in filenames:
            dirnames.clear()
            continue

        package_name = convert_folder_to_package(dirpath)

        if package_name in ignores:
            dirnames.clear()
            continue

        yield package_name

        for filename in filenames:
            if not filename.endswith('.py'):
                continue
            if filename == '__init__.py':
                continue

            module_name = '.'.join([convert_folder_to_package(dirpath), filename[:-3]])

            if module_name not in ignores:
                yield module_name

        try:
            dirnames.remove('__pycache__')
        except ValueError:
            pass


def relative_path(base_filepath, *subpaths):
    return os.path.abspath(os.path.join(os.path.dirname(base_filepath), *subpaths))


@pytest.mark.parametrize(
    'module_name', list_packages_in_folder(
        relative_path(__file__, '..', 'lookups'),
        ignores=[
        ]
    )
)
def test_imports(module_name):
    assert importlib.import_module(module_name)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
import typing as T
from weakref import WeakKeyDictionary, CallableProxyType, proxy as weakproxy

# Third-party imports
from observable import Observable

# Local imports


class WeakObservable(Observable):
    """An Observable with events that are objects instead of str, and we only keep weakrefs on them.
    Handlers will also be saved as proxies."""

    def __init__(self) -> None:
        super().__init__()
        self._events: WeakKeyDictionary[T.Any, T.List[CallableProxyType]] = WeakKeyDictionary()

    def on(  # pylint: disable=invalid-name
        self, event: T.Any, *handlers: T.Callable
    ) -> T.Callable:
        """Registers one or more handlers to a specified event.
        This method may as well be used as a decorator for the handler."""

        def _on_wrapper(*handlers: T.Callable) -> T.Callable:
            """wrapper for on decorator"""
            if event not in self._events:
                self._events[event] = []
            for handler in handlers:
                self._events[event].append(weakproxy(handler))
            return handlers[0]

        if handlers:
            return _on_wrapper(*handlers)
        return _on_wrapper



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from collections.abc import Hashable, Sequence, MutableSequence, Set, MutableSet

# Third-party imports
import pytest

# Local imports
from lookups import singleton
from .tools import TestParentObject, TestChildObject, TestOtherObject


MEMBER_FIXTURES = [
    # None
    pytest.param(None, None, object, marks=pytest.mark.xfail),
    # object
    (object(), None, object),
    (object(), 'obj', object),
    pytest.param(object(), None, TestOtherObject, marks=pytest.mark.xfail),
    # TestParentObject
    (TestParentObject(), None, object),
    (TestParentObject(), 'parent', object),
    (TestParentObject(), None, TestParentObject),
    (TestParentObject(), 'parent', TestParentObject),
    pytest.param(TestParentObject(), None, TestChildObject, marks=pytest.mark.xfail),
    pytest.param(TestParentObject(), 'parent', TestChildObject, marks=pytest.mark.xfail),
    pytest.param(TestParentObject(), None, TestOtherObject, marks=pytest.mark.xfail),
    pytest.param(TestParentObject(), 'parent', TestOtherObject, marks=pytest.mark.xfail),
    # TestChildObject
    (TestChildObject(), None, object),
    (TestChildObject(), 'child', object),
    (TestChildObject(), None, TestParentObject),
    (TestChildObject(), 'child', TestParentObject),
    (TestChildObject(), None, TestChildObject),
    (TestChildObject(), 'child', TestChildObject),
    pytest.param(TestChildObject(), None, TestOtherObject, marks=pytest.mark.xfail),
    pytest.param(TestChildObject(), 'child', TestOtherObject, marks=pytest.mark.xfail),
]


def check_all_instances(member, all_instances):
    assert isinstance(all_instances, Sequence)
    assert not isinstance(all_instances, MutableSequence)
    assert len(all_instances) == 1
    assert all_instances == (member, )


def check_item(member, id_, item):
    assert item is not None
    assert isinstance(item, Hashable)

    assert item.get_display_name()
    if id_ is not None:
        assert item.get_id() == id_
    else:
        assert item.get_id()

    assert item.get_instance() is member

    assert issubclass(item.get_type(), type(member))


def test_instantiation():
    assert singleton(object(), None)


@pytest.mark.parametrize('member, id_, search', MEMBER_FIXTURES)
def test_lookup(member, id_, search):
    lookup = singleton(member, id_)
    assert lookup.lookup(search) is member


@pytest.mark.parametrize('member, id_, search', MEMBER_FIXTURES)
def test_lookup_item(member, id_, search):
    lookup = singleton(member, id_)

    item = lookup.lookup_item(search)
    check_item(member, id_, item)
    assert item == lookup.lookup_item(search)


@pytest.mark.parametrize('member, id_, search', MEMBER_FIXTURES)
def test_lookup_all(member, id_, search):
    lookup = singleton(member, id_)

    all_instances = lookup.lookup_all(search)
    check_all_instances(member, all_instances)


@pytest.mark.parametrize('member, id_, search', MEMBER_FIXTURES)
def test_lookup_result(member, id_, search):
    lookup = singleton(member, id_)

    result = lookup.lookup_result(search)
    assert result

    all_classes = result.all_classes()
    assert isinstance(all_classes, Set)
    assert not isinstance(all_classes, MutableSet)
    assert len(all_classes) == 1
    assert next(iter(all_classes)) is type(member)

    all_instances = result.all_instances()
    check_all_instances(member, all_instances)

    all_items = result.all_items()
    assert isinstance(all_items, Sequence)
    assert not isinstance(all_items, MutableSequence)
    assert len(all_items) == 1
    check_item(member, id_, all_items[0])
    assert all_items[0] == lookup.lookup_item(search)


@pytest.mark.parametrize('member, id_, search', MEMBER_FIXTURES)
def test_listeners(member, id_, search):
    lookup = singleton(member, id_)

    result = lookup.lookup_result(search)

    def call_me_back(result):
        pass

    result.add_lookup_listener(call_me_back)
    result.remove_lookup_listener(call_me_back)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations  # noqa: F407

# System imports
import logging
from abc import ABC, abstractmethod
from typing import Sequence, AbstractSet, Type, Optional, Callable

# Third-party imports

# Local imports


_logger = logging.getLogger(__name__)


class Lookup(ABC):
    '''
    A general registry permitting clients to find instances of services (implementation of a given
    interface).

    This class is inspired Netbeans Platform lookup mechanism. The Lookup API concentrates on the
    lookup, not on the registration.
    '''

    @classmethod
    def get_default(cls) -> Lookup:
        '''
        Method to obtain the global lookup in the whole system.

        The actual returned implementation can be different in different systems, but the default
        one is based on lookup.Lookups.???.

        :return: The global lookup in the system
        :rtype: Lookup
        '''
        # Temporary solution
        from .generic_lookup import GenericLookup
        from .instance_content import InstanceContent
        return GenericLookup(InstanceContent())

    @abstractmethod
    def lookup(self, cls: Type[object]) -> Optional[object]:
        '''
        Look up an object matching a given type.

        If more than one object matches, the first will be returned.
        The template class may be a class or a type.
        The instance is guaranteed to be of the same type.

        :param cls: Class or type of the object we are searching for.
        :return: An object implementing the given type or None if no such implementation is found.
        '''

    @abstractmethod
    def lookup_result(self, cls: Type[object]) -> Result:
        '''
        Find a result corresponding to a given class.

        Equivalent to calling Lookup.Template but slightly more convenient.
        Subclasses may override this method to produce the same semantics more efficiently.

        :param cls: Class or type of the objects we are searching for.
        :return: A live object representing instances of that type.
        '''

    def lookup_item(self, cls: Type[object]) -> Optional[Item]:
        '''
        Look up the first item matching a given class.

        Includes not only the instance but other associated information.

        :param cls: Class or type of the objects we are searching for.
        :return: A matching item or None.
        '''
        result = self.lookup_result(cls)
        it = iter(result.all_items())
        return next(it, None)

    def lookup_all(self, cls: Type[object]) -> Sequence[object]:
        '''
        Find all instances corresponding to a given class.

        Equivalent to calling lookup_result().all_instances(), but slightly more convenient.
        Subclasses may override this method to produce the same semantics more efficiently.

        > for service in Lookup.get_default().lookup_all(type(MyService)):
        >     service.use_me()

        :param cls: Class or type of the objects we are searching for.
        :return: All currently available instances of that type.
        '''
        return self.lookup_result(cls).all_instances()


class Item(ABC):
    '''
    A single item in a lookup result.

    This wrapper provides unified access to not just the instance, but its class, a possible
    persistent identifier, and so on.
    '''

    @abstractmethod
    def get_display_name(self) -> str:
        '''
        Get a human presentable name for the item.

        This might be used when summarizing all the items found in a lookup result
        in some part of a GUI.

        :return: The string suitable for presenting the object to a user.
        '''

    @abstractmethod
    def get_id(self) -> str:
        '''
        Get a persistent identifier for the item.

        This identifier should uniquely represent the item within its containing lookup (and if
        possible within the global lookup as a whole). For example, it might represent the source of
        the instance as a file name. The ID may be persisted and in a later session used to find the
        same instance as was encountered earlier, by means of passing it into a lookup template.

        :return: A string ID of the item.
        '''

    @abstractmethod
    def get_instance(self) -> Optional[object]:
        '''
        Get the instance itself.

        :return: The instance or None if the instance cannot be created.
        '''

    @abstractmethod
    def get_type(self) -> Type[object]:
        '''
        Get the implementing class of the instance.

        :return: The class of the item.
        '''

    def __str__(self) -> str:
        '''Show ID for debugging.'''
        return self.get_id()


class Result(ABC):
    '''
    Result of a lookup request.

    Allows access to all matching instances at once.
    Also permits listening to changes in the result.
    Result can contain duplicate items.
    '''

    @abstractmethod
    def add_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        '''
        Registers a listener that is invoked when there is a possible change in this result.

        LookupListener protocol is to implement the following method:

            def result_changed(self, result: Result) -> None: ...

        Remember to keep a strong reference to the object you are attaching listener to.

        :param listener: The listener to add.
        '''

    @abstractmethod
    def remove_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        '''
        Unregisters a listener previously added.

        :param listener: The listener to remove.
        '''

    @abstractmethod
    def all_classes(self) -> AbstractSet[Type[object]]:
        '''
        Get all classes represented in the result.

        That is, the set of concrete classes used by instances present in the result.
        All duplicate classes will be omitted.

        :return: Immutable set of class objects.
        '''

    @abstractmethod
    def all_instances(self) -> Sequence[object]:
        '''
        Get all instances in the result.

        :return: Immutable sequence of all instances.
        '''

    @abstractmethod
    def all_items(self) -> Sequence[Item]:
        '''
        Get all registered items.

        This should include all pairs of instances together with their classes, IDs, and so on.

        :return: Immutable sequence of Item.
        '''


class LookupProvider(ABC):
    '''
    Classes implementing interface lookup.Provider are capable of
    and willing to provide a lookup (usually bound to the object).
    '''

    @abstractmethod
    def get_lookup(self) -> Lookup:
        '''
        Returns lookup associated with the object.

        :return: Fully initialized lookup instance provided by this object
        '''



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
import codecs
import json
import os
import setuptools

# Third-party imports

# Local imports


ROOT_PATH = os.path.abspath(os.path.dirname(__file__))


def read_text(*paths):
    '''Returns text content of a file.'''
    with codecs.open(os.path.join(ROOT_PATH, *paths), 'r', 'utf-8') as f:
        return f.read()


def read_py(*paths):
    '''Executes a Python file and returns its globals namespace.'''
    into = {}
    exec(read_text(*paths), into)
    return into


def pipfile_requirements(section, with_version=False):
    '''Extract requirements from Pipfile.lock'''
    pipfile = json.loads(read_text('Pipfile.lock'))

    if with_version:
        return [
            package + detail.get('version', '')
            for package, detail in pipfile[section].items()
        ]
    else:
        return [
            package
            for package, detail in pipfile[section].items()
        ]


about = read_py('lookups', '__about__.py')
setup_args = dict(
    # Meta info
    name=about['__title__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__email__'],
    license=about['__license__'],
    description=about['__summary__'],
    long_description='\n\n'.join([
        read_text('README.md'),
        read_text('AUTHORS.md'),
        read_text('CHANGELOG.md'),
    ]),
    long_description_content_type='text/markdown',
    url=about['__uri__'],

    # Code and dependencies
    packages=setuptools.find_packages(exclude=['tests']),
    # install_requires=pipfile_requirements('default'),  # , with_version=True),
    install_requires=[
        'observable',
    ],
    tests_require=pipfile_requirements('develop'),
    python_requires='~=3.7',

    # Additional files
    package_data=dict(lookups=['py.typed']),

    # Classifiers and Keywords
    classifiers=(  # https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Object Brokering',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Typing :: Typed',
    ),
    keywords='lookup lookups dci',

    # Other options
    zip_safe=False,  # Because mypy cannot analyse a zip
)

if __name__ == '__main__':
    setuptools.setup(**setup_args)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from typing import Sequence, AbstractSet, Type, Optional, Callable

# Third-party imports

# Local imports
from . import lookups
from .lookup import Lookup, Result, Item


class SimpleLookup(Lookup):
    '''
    Simple lookup implementation. It can be used to create temporary lookups that do not change over
    time. The result stores references to all objects passed in the constructor. Those objecst are
    the only ones returned as result.
    '''

    def __init__(self, *instances: object) -> None:
        '''
        Creates new SimpleLookup object with supplied instances parameter.

        :param instances: Instances to be used to return from the lookup.
        '''
        self.all_items: Sequence[Item] = tuple(
            lookups.LookupItem(instance) for instance in instances)

    def lookup(self, cls: Type[object]) -> Optional[object]:
        for item in self.all_items:
            if issubclass(item.get_type(), cls):
                return item.get_instance()
        else:
            return None

    def lookup_result(self, cls: Type[object]) -> Result:
        return SimpleResult(self, cls)


class SimpleResult(Result):
    '''
    Result used in SimpleLookup. It holds a reference to the collection passed in constructor.
    As the contents of this lookup result never changes the add_lookup_listener() and
    remove_lookup_listener do not do anything.
    '''

    def __init__(self, simple_lookup: SimpleLookup, cls: Type[object]) -> None:
        self.lookup = simple_lookup
        self.cls = cls
        self._items: Optional[Sequence[Item]] = None

    def add_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def remove_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def all_classes(self) -> AbstractSet[Type[object]]:
        return frozenset(
            item.get_type() for item in self.all_items()
        )

    def all_instances(self) -> Sequence[object]:
        return tuple(filter(
            None,
            (item.get_instance() for item in self.all_items())
        ))

    def all_items(self) -> Sequence[Item]:
        if self._items is None:
            self._items = tuple(
                item for item in self.lookup.all_items
                if issubclass(item.get_type(), self.cls)
            )

        return self._items



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations  # noqa: F407

# System imports
from abc import ABC, abstractmethod
from concurrent.futures import Executor
from contextlib import contextmanager
from threading import RLock
from typing import (
    Iterable, Collection,
    MutableSequence, Sequence,
    AbstractSet, Set,
    Type, Optional,
    Iterator, Tuple,
    Callable,
)

# Third-party imports

# Local imports
from .lookup import Lookup, Item, Result
from .weak_observable import WeakObservable


class GenericLookup(Lookup):
    '''
    Implementation of the lookup from OpenAPIs that is based on the insertion of Item.

    This class should provide the default way of how to store (Class, Object) pairs in the lookups.
    It offers protected methods for subclasses to register the pairs.

    Note:
    - GenericLookup is the API for users of the lookup.
    - Content is the API for owner/creator of the lookup.
    - Storage is the actual internal place where (Class, Object) pairs are stored.
    '''

    def __init__(self, content: Content) -> None:
        '''
        Constructor to create this lookup and associate it with given Content.

        The content then allows the creator to invoke protected methods which are not accessible for
        any other user of the lookup.

        :param content: The content to associate with.
        '''
        self._storage: Optional[Storage] = None
        self._storage_lock = RLock()
        self._storage_is_used = False
        self._observers = WeakObservable()

        content._attach(self)

    @contextmanager
    def _storage_for_lookup(self) -> Iterator[Storage]:
        from .set_storage import SetStorage

        with self._storage_lock:
            # Enter storage
            if self._storage is None:
                self._storage = SetStorage()
                self._initialise()

            yield self._storage

    @contextmanager
    def _storage_for_modification(
            self, ensure: int, notify_in: Executor = None) -> Iterator[Transaction]:

        with self._storage_for_lookup() as storage:
            if self._storage_is_used:
                raise RuntimeError('You are trying to modify a lookup from a lookup query!')
            self._storage_is_used = True

            try:
                transaction = storage.begin_transaction(ensure)
                yield transaction
                to_notify = storage.end_transaction(transaction)
            finally:
                # Exit storage
                self._storage_is_used = False

            self._notify_in(notify_in, to_notify)

    def _initialise(self) -> None:
        '''Method for subclasses to initialize themselves.'''

    def _before_lookup(self, cls: Type[object]) -> None:
        '''
        Notifies subclasses that a query is about to be processed.

        :param cls: Class or type of the objects searched for.
        '''

    def _add_pair(self, pair: Pair, notify_in: Executor = None) -> bool:
        '''
        The method to add instance to the lookup with.

        :param pair: Class/instance pair.
        :param notify_in: The executor that will handle the notification of events.
        :return: True if the pair has been added for the first time or False if some other pair
        equal to this one already existed in the lookup
        '''
        with self._storage_for_modification(-2, notify_in) as transaction:
            return transaction.add(pair)

    def _remove_pair(self, pair: Pair, notify_in: Executor = None) -> None:
        '''
        Remove instance.

        :param pair: Class/instance pair.
        :param notify_in: The executor that will handle the notification of events.
        '''
        with self._storage_for_modification(-1, notify_in) as transaction:
            transaction.remove(pair)

    def _set_pairs(self, pairs: Collection[Pair], notify_in: Executor = None) -> None:
        '''
        Changes all pairs in the lookup to new values, notifies listeners using provided executor.

        :param pairs: The collection of class/instance pairs.
        :param notify_in: The executor that will handle the notification of events.
        '''
        with self._storage_for_modification(len(pairs), notify_in) as transaction:
            transaction.set_all(pairs)

    def _notify_in(self, notify_in: Optional[Executor], listeners: Iterable[Result]) -> None:
        if not notify_in:
            for result in listeners:
                self._observers.trigger(result, result)
        else:
            notify_in.map(self._observers.trigger, listeners, listeners)

    def lookup(self, cls: Type[object]) -> Optional[object]:
        item = self.lookup_item(cls)
        return item.get_instance() if item is not None else None

    def lookup_result(self, cls: Type[object]) -> Result:
        self._before_lookup(cls)
        with self._storage_for_lookup() as storage:
            result = storage.find_result(cls)
            if result:
                return result

            result = GLResult(self, cls)
            storage.register_result(result)

            return result

    def lookup_item(self, cls: Type[object]) -> Optional[Item]:
        self._before_lookup(cls)
        with self._storage_for_lookup() as storage:
            for pair in storage.lookup(cls):
                return pair
            else:
                return None

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(<content>)'


# For Java folks ;)
AbstractLookup = GenericLookup


class Pair(Item):
    '''
    Extension to the default lookup item that offers additional information for the data structures
    used in GenericLookup.
    '''


class GLResult(Result):

    def __init__(self, lookup: GenericLookup, cls: Type[object]) -> None:
        self._lookup = lookup
        self._cls = cls
        self._listeners = None

        self._classes_cache: Optional[AbstractSet[Type[object]]] = None
        self._items_cache: Optional[Sequence[Item]] = None
        self._instances_cache: Optional[Sequence[object]] = None

    def clear_cache(self) -> None:
        '''To be called when a result is affected by a change during a transaction.'''
        self._classes_cache = None
        self._items_cache = None
        self._instances_cache = None

    def add_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        self._lookup._observers.on(self, listener)

    def remove_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        self._lookup._observers.off(self, listener)

    def all_classes(self) -> AbstractSet[Type[object]]:
        self._lookup._before_lookup(self._cls)

        if self._classes_cache is not None:
            return self._classes_cache

        classes = frozenset([
            item.get_type()
            for item in self._all_items_without_before_lookup()
        ])

        self._classes_cache = classes
        return classes

    def all_instances(self) -> Sequence[object]:
        self._lookup._before_lookup(self._cls)

        if self._instances_cache:
            return self._instances_cache

        instances = tuple([
            item.get_instance()
            for item in self._all_items_without_before_lookup()
            if issubclass(item.get_type(), self._cls)
        ])

        self._instances_cache = instances
        return instances

    def all_items(self) -> Sequence[Item]:
        self._lookup._before_lookup(self._cls)
        return self._all_items_without_before_lookup()

    def _all_items_without_before_lookup(self) -> Sequence[Item]:
        if self._items_cache:
            return self._items_cache

        with self._lookup._storage_for_lookup() as storage:
            pairs = tuple(storage.lookup(self._cls))
            self._items_cache = pairs
            return pairs

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._lookup!r}, {self._cls!r})'


class Content:
    '''
    A class that can be used by the creator of the GenericLookup to control its content (a kind of
    Privileged API giving creator of the lookup more rights than subsequent users of the lookup).

    Note that a Content does not store any lookup item!
    The Storage instanciated by GenericLookup does.
    '''

    def __init__(self, notify_in: Executor = None) -> None:
        '''
        Creates a content associated with an executor to handle dispatch of changes.

        :param notify_in: The executor to notify changes in (ie. to listeners of lookup results).
        '''
        self._notify_in: Optional[Executor] = notify_in  # Do not serialize
        self._abstract_lookup: Optional[GenericLookup] = None
        self._early_pairs: MutableSequence[Pair] = []

    def _attach(self, lookup: GenericLookup) -> None:
        '''
        A lookup attaches to this object.
        '''
        if self._abstract_lookup is None:
            self._abstract_lookup = lookup

            if self._early_pairs:
                self._set_pairs(self._early_pairs)
                self._early_pairs = []

        else:
            raise RuntimeError(
                f'Trying to use content for {lookup!r} '
                f'but it is already used for {self._abstract_lookup!r}'
            )

    def _add_pair(self, pair: Pair) -> bool:
        '''
        The method to add a pair to the associated GenericLookup.

        Preferably call this method when lookup is already associated with this content (association
        is done by passing this object to some GenericLookup's constructor once).

        :param pair: Class/instance pair.
        :return: True if the pair has been added for the first time or False if some other pair
        equal to this one already existed in the lookup
        '''
        if self._abstract_lookup:
            return self._abstract_lookup._add_pair(pair, self._notify_in)
        else:
            present = pair in self._early_pairs
            self._early_pairs.append(pair)
            return present

    def _remove_pair(self, pair: Pair) -> None:
        '''
        Remove instance.

        :param pair: Class/instance pair.
        '''
        if self._abstract_lookup:
            self._abstract_lookup._remove_pair(pair, self._notify_in)
        else:
            self._early_pairs.remove(pair)

    def _set_pairs(self, pairs: Collection[Pair]) -> None:
        '''
        Changes all pairs in the lookup to new values.

        :param pairs: The collection of class/instance pairs.
        '''
        if self._abstract_lookup:
            self._abstract_lookup._set_pairs(pairs, self._notify_in)
        else:
            self._early_pairs = list(pairs)


class Storage(ABC):
    '''Storage to keep the internal structure of Pairs and to answer different queries.'''

    @abstractmethod
    def begin_transaction(self, ensure: int) -> Transaction:
        '''
        Initializes a modification operation by creating an object that will be passsed to all add,
        remove, retainAll methods and should collect enough information about the change to notify
        listeners about the transaction later.

        :param ensure: The amount of items that will appear in the storage after the modifications:
         * -1 == remove one
         * -2 == add one
         * >= 0: the amount of objects at the end
        :return: A transaction object
        '''

    @abstractmethod
    def end_transaction(self, transaction: Transaction) -> Iterable[GLResult]:
        '''
        Collects all affected results that were modified in the given transaction.

        :param transaction: The transaction object.
        :return: The results affected by a change.
        '''

    @abstractmethod
    def lookup(self, cls: Type[object]) -> Iterable[Pair]:
        '''
        Queries for instances of given class.

        :param cls: The class to search for.
        :return: Iterable of Item
        '''

    @abstractmethod
    def register_result(self, result: GLResult) -> None:
        '''
        Registers a result with the storage.

        :param result: The new result to remember.
        '''

    @abstractmethod
    def find_result(self, cls: Type[object]) -> Optional[GLResult]: ...


class Transaction(ABC):
    '''Keeps track of changes happening in a Storage'''

    @abstractmethod
    def add(self, pair: Pair) -> bool:
        '''
        Adds an item into the storage.

        :param item: Item to add.
        :return: True if the Item has been added for the first time or False if some other item
        equal to this one already existed in the lookup
        '''

    @abstractmethod
    def remove(self, pair: Pair) -> None:
        '''
        Removes an item.

        :param item: Item to remove.
        '''

    @abstractmethod
    def set_all(self, pairs: Collection[Pair]) -> None:
        '''
        Changes all pairs to new values.

        :param pairs: The collection of class/instance pairs.
        '''

    @abstractmethod
    def new_content(self, prev: Collection[Pair]) -> Tuple[Collection[Pair], Set[Pair]]:
        '''
        Return the new content once a transaction is finished.

        :param prev: The previous content.
        :return: Tuple of new content, set of changes.
        '''



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from collections.abc import Sequence, MutableSequence, Set, MutableSet

# Third-party imports

# Local imports
from lookups.lookups import EmptyLookup


def check_all_instances(all_instances):
    assert isinstance(all_instances, Sequence)
    assert not isinstance(all_instances, MutableSequence)
    assert len(all_instances) == 0


def test_instantiation():
    assert EmptyLookup()


def test_lookup():
    lookup = EmptyLookup()
    assert lookup.lookup(object) is None


def test_lookup_item():
    lookup = EmptyLookup()

    item = lookup.lookup_item(object)
    assert item is None


def test_lookup_all():
    lookup = EmptyLookup()

    all_instances = lookup.lookup_all(object)
    check_all_instances(all_instances)


def test_lookup_result():
    lookup = EmptyLookup()

    result = lookup.lookup_result(object)
    assert result

    all_classes = result.all_classes()
    assert isinstance(all_classes, Set)
    assert not isinstance(all_classes, MutableSet)
    assert len(all_classes) == 0

    all_instances = result.all_instances()
    check_all_instances(all_instances)

    all_items = result.all_items()
    assert isinstance(all_items, Sequence)
    assert not isinstance(all_items, MutableSequence)
    assert len(all_items) == 0


def test_listeners():
    lookup = EmptyLookup()

    result = lookup.lookup_result(object)

    def call_me_back(result):
        pass

    result.add_lookup_listener(call_me_back)
    result.remove_lookup_listener(call_me_back)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports
from typing import Sequence, AbstractSet, Type, Optional, Callable

# Third-party imports

# Local imports
from . import lookups
from .lookup import Lookup, Result, Item


class SingletonLookup(Lookup):
    '''
    Unmodifiable lookup that contains just one fixed object.
    '''

    def __init__(self, member: object, id_: str = None) -> None:
        '''
        :param member: The only fixed instance in this lookup.
        :param id_: Persistent identifier for member.
        '''
        if member is None:
            raise ValueError('None cannot be a lookup member')

        self._member = member
        self._id = id_

    def lookup(self, cls: Type[object]) -> Optional[object]:
        if isinstance(self._member, cls):
            return self._member
        else:
            return None

    def lookup_result(self, cls: Type[object]) -> Result:
        item = self.lookup_item(cls)
        if item is not None:
            return SingletonResult(item)
        else:
            return lookups.EmptyLookup().lookup_result(cls)

    def lookup_item(self, cls: Type[object]) -> Optional[Item]:
        if isinstance(self._member, cls):
            return lookups.LookupItem(self._member, self._id)
        else:
            return None

    def lookup_all(self, cls: Type[object]) -> Sequence[object]:
        if isinstance(self._member, cls):
            return (self._member, )
        else:
            return tuple()

    def __str__(self) -> str:
        return 'SingletonLookup[%s]' % str(self._member)


class SingletonResult(Result):

    def __init__(self, item: Item) -> None:
        self._item = item

    def add_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def remove_lookup_listener(self, listener: Callable[[Result], None]) -> None:
        pass

    def all_classes(self) -> AbstractSet[Type[object]]:
        return frozenset((self._item.get_type(), ))

    def all_instances(self) -> Sequence[object]:
        instance = self._item.get_instance()
        if instance is not None:
            return self._item.get_instance(),
        else:
            return tuple()

    def all_items(self) -> Sequence[Item]:
        return self._item,



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports

# Third-party imports

# Local imports

__all__ = [
    '__title__', '__summary__', '__uri__', '__version__',
    '__author__', '__email__', '__license__', '__copyright__',
]

__title__ = 'lookups'
__summary__ = 'DCI lookups in Python (inspired by Netbeans Platform Lookups API)'
__uri__ = 'https://github.com/AxelVoitier/lookups'

__version__ = '0.2.0'

__author__ = 'Axel Voitier'
__email__ = 'axel.voitier@gmail.com'

__license__ = 'MPL-2.0'
__copyright__ = 'Copyright 2019 Lookups contributors (see AUTHORS.md)'
